""" Model definitions for the quiz """

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models


class CriterionRating(models.Model):
    """
    This is the relation between Sport and Criterion.
    You can use it to add a rating for a specific criterion to a sport.
    To see it's usage check Sport.rate() and Sport.get_rating()
    """

    rating = models.IntegerField(
        validators=[MaxValueValidator(10), MinValueValidator(1)]
    )
    criterion = models.ForeignKey("Criterion", on_delete=models.CASCADE)
    sport = models.ForeignKey("Sport", on_delete=models.CASCADE)

    def __str__(self):
        return str(self.sport) + " - " + str(self.criterion) + ": " + str(self.rating)


class Sport(models.Model):
    """
    Defines a Sport with name, url that leads to the booking page.
    A sport includes ratings for all criterions.
    (e.g. How much it corresponds to the criterion "Martial Arts")
    """

    name = models.TextField()
    url = models.URLField()
    criteria_ratings = models.ManyToManyField("Criterion", through="CriterionRating")

    def __str__(self):
        return self.name

    def rate(self, criterion, rating):
        """Defines how much (rating) the sport meets the given criterion"""
        rating_obj, _ = CriterionRating.objects.get_or_create(
            sport=self, criterion=criterion, defaults={"rating": rating}
        )
        rating_obj.rating = rating
        rating_obj.save()
        return rating_obj

    def get_rating(self, criterion):
        """Returns how much the sport meets the given criterion"""
        criterion_rating = CriterionRating.objects.get(sport=self, criterion=criterion)
        return criterion_rating.rating


class Criterion(models.Model):
    """
    Defines a Sport property that is used a a criterion for our quiz.
    (e.g. Individual or Team sport)
    """

    name = models.TextField()

    def __str__(self):
        return self.name


class CallToMove(models.Model):
    """Defines text and image that are used to show a call to move between questions"""

    text = models.TextField()
    image = models.ImageField(null=True, max_length=200)

    def __str__(self):
        return self.text


class KnowledgeSnack(models.Model):
    """Defines text and image that are used to show a KnowledgeSnack between questions"""

    text = models.TextField()
    image = models.ImageField(null=True, max_length=200)

    def __str__(self):
        return self.text


class Question(models.Model):
    """Defines a Question that is assigned to exactly one Criterion"""

    text = models.TextField()
    criterion = models.OneToOneField(
        Criterion, on_delete=models.CASCADE, primary_key=True
    )

    def __str__(self):
        return self.text


class QuestionOrderManager(models.Manager):
    """
    Manages Handling of QuestionOrder Objects
    Includes create_entry_at_end, which handles Creation of new Entries
    and delete_entry, which deletes one entry and changes the order_id of the following entries
    """

    def create_entry_at_end(self, type_of_slot, question_id=None):
        """
        Creates a new OrderEntry at the end of the current Order (so the object gets given the highest order_id)
        """

        if type_of_slot == "question" and question_id is None:
            raise ValueError(
                "A Question ID must be given if the Type of the Slot is Question"
            )
        elif type_of_slot not in [
            "question",
            "snack",
            "activity",
        ]:
            raise ValueError(
                f'{type_of_slot} not in valid choice list ["question", "snack", "activity"]'
            )

        entry = QuestionOrder()

        entry.type_of_slot = type_of_slot

        # If question_id is None, fill it out as -1, else take the question_id value
        if entry.type_of_slot == "question":
            entry.question_id = question_id

        # If no Entry exists, highest_current_order is 0, otherwise highest order_id
        if self.count() == 0:
            highest_current_order = 0
        else:
            # "latest" returns highest value (normally used for dates, I believe)
            highest_current_order = self.latest("order_id").order_id

        entry.order_id = highest_current_order + 1

        entry.save()

        return entry

    def delete_entry(self, given_order_id):
        """
        Delete an Entry in the QuestionOrder Database and decrement every object that had a larger order_id than the one deleted
        """

        # Delete must be called on an instance, not a queryset
        entry = self.get(order_id=given_order_id)
        entry.delete()

        larger_entries = self.filter(order_id__gte=given_order_id)

        # The Primary Key of an Object cannot be changed, really, instead a new Object is created when the PK changes
        # This means that we need to delete the original Object so that we only have the new pk

        for entry in larger_entries.iterator():
            entry.order_id = entry.order_id - 1
            entry.save()

    def delete_entry_by_question_id(self, given_question_id):
        """
        Delete an Entry in the QuestionOrder DB by the Question_ID
        """

        queryset = self.filter(question_id=given_question_id)

        if queryset.count() == 0:
            # If the Question doesn't exist in the Order, we don't need to do anything
            return

        # If the question is in the Order more than once for some reason, sweat not, we will delete them all
        for entry in queryset.iterator():
            self.delete_entry(entry.order_id)


class QuestionOrder(models.Model):
    """
    Defines the order of the Questions, Snacks, and Activities given in the Quiz
    Default Choice for type_of_slot is "snack" instead of "question",
    because "snack" doesn't need a question_id and is therefore a safer choice
    """

    objects = QuestionOrderManager()

    order_id = models.IntegerField(null=True)
    type_of_slot = models.TextField(
        choices=[
            ("question", "question"),
            ("snack", "snack"),
            ("activity", "activity"),
        ],
        default="snack",
    )
    question_id = models.IntegerField(default=-1)

    def __str__(self):
        return f"Entry {self.order_id}: {self.type_of_slot}"
